import { Component, OnInit } from '@angular/core';
import { WorkOrderService } from '../services/work-order.service';
import { WorkOrder } from '../models/work-order.model';
import { Technician } from '../models/technician.model'; 
import { FormsModule } from '@angular/forms';
import { CommonModule } from '@angular/common';
import { NgbModal } from '@ng-bootstrap/ng-bootstrap';
import { NewWorkOrderModalComponent } from '../new-work-order-modal/new-work-order-modal.component';
import { HttpClient } from '@angular/common/http';

@Component({
  selector: 'app-work-orders',
  standalone: true,
  templateUrl: './work-orders.component.html',
  styleUrls: ['./work-orders.component.css'],
  imports: [CommonModule,FormsModule]
})
export class WorkOrdersComponent implements OnInit {
  selectedStatus: string = 'All';
  statuses: string[] = ['New', 'In Progress', 'Completed'];
  workOrders: any[] = [];
  selectedWorkOrder: WorkOrder = {} as WorkOrder;
  technicians: any[] = [];
  filteredWorkOrders: any[] = [];
  newWorkOrder: WorkOrder = {
    woNum: 0,
    email: '',
    contactName: '',
    contactNumber: '',
    problem: '',
    technician: 1,
    technicianComments: '',
    dateReceived: '',
    $id: '',
    status: 'open' 
  };

  constructor(
    private modalService: NgbModal,
    private workOrderService: WorkOrderService,
    private http: HttpClient
  ) { }

  ngOnInit(): void {
    this.http.get<any>('https://lc-code-test-api.azurewebsites.net/api/workorders')
      .subscribe(
        (response) => {
          console.log('API Response:', response);
          if (response && response.$values) {
            this.workOrders = response.$values;
            this.workOrders.forEach(workOrder => {
              if (workOrder.technician) {
                console.log(workOrder.technician);  
              } else {
                console.log("No technician data for work order", workOrder.woNum);
              }
            });
  
            this.filterWorkOrders();
          } else {
            console.error('Error: missing work orders.');
          }
        },
        (error) => {
          console.error('Error fetching work orders or technicians.', error);
        }
      );
  }
  
  sanitizeDate(date: any): any {
    const validDate = new Date(date);
    return isNaN(validDate.getTime()) ? null : validDate;
  }
  
  loadWorkOrdersByStatus(status: string): void {
    this.workOrderService.getWorkOrdersByStatus(status).subscribe(
      (data) => {
        this.workOrders = data;
      },
      (error) => {
        console.error('Error loading work orders:', error);
      }
    );
  }

  loadWorkOrders(): void {
    this.workOrderService.getWorkOrders().subscribe(
      (data) => {
        this.workOrders = data; 
      },
      (error) => {
        console.error('Error fetching work orders', error);
      }
    );
  }

  loadTechnicians(): void {
    this.workOrderService.getTechnicians().subscribe(
      (data: any) => {
        this.technicians = data.technicians;
      },
      (error) => {
        console.error('Error loading technicians:', error);
      }
    );
  }

  onStatusChange(): void {
    console.log('Selected status:', this.selectedStatus);
    this.loadWorkOrdersByStatus(this.selectedStatus);
  }

  getWorkOrders() {
    this.workOrderService.getWorkOrders().subscribe(
      (data: any) => {
        this.workOrders = data.$values;  
        this.filterWorkOrders();  
      },
      (error) => {
        console.error('Error fetching work orders', error);
      }
    );
  }

  filterWorkOrders() {
    if (this.selectedStatus === 'All') {
      this.filteredWorkOrders = [...this.workOrders];
    } else {
      this.filteredWorkOrders = this.workOrders.filter((workOrder: WorkOrder) => workOrder.status === this.selectedStatus);
    }
  }
  
  openModal(): void {
    const modalRef = this.modalService.open(NewWorkOrderModalComponent);
    modalRef.componentInstance.technicians = this.technicians;  
  }

  getTechnicians() {
    this.workOrderService.getTechnicians().subscribe(technicians => {
      this.technicians = technicians;
    });
  }
  
  onSubmit() {
    if (!this.newWorkOrder.contactName) {
      this.newWorkOrder.contactName = 'Default Name';
    }
  
    if (!this.newWorkOrder.email) {
      this.newWorkOrder.email = 'default@lakecountyfl.gov';
    }
  
    if (!this.newWorkOrder.problem) {
      this.newWorkOrder.problem = 'No description provided';
    }
  
    this.newWorkOrder.dateReceived = new Date().toISOString();
  
    this.workOrderService.createWorkOrder(this.newWorkOrder).subscribe(response => {
      console.log('Work order created:', response);
      this.workOrders.push(response); 
    }, error => {
      console.error('Error creating work order', error);
    });
  }
  
  openNewWorkOrderModal(): void {
    const modalRef = this.modalService.open(NewWorkOrderModalComponent, {
      size: 'lg',
      backdrop: 'static'
    });

    modalRef.componentInstance.technicians = this.technicians;

    modalRef.result.then(
      (result) => {
        console.log('Modal closed with result:', result);
      },
      (reason) => {
        console.log('Modal dismissed with reason:', reason);
      }
    );
  }

  getWorkOrdersByTechnician(technicianId: number): void {
    this.workOrderService.getWorkOrdersByTechnician(technicianId)
      .subscribe(
        (data) => {
          this.workOrders = data;
          console.log('Work Orders:', this.workOrders);
        },
        (error) => {
          console.error('Error fetching work orders:', error);
        }
      );
  }
}